#include <conio.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <peekpoke.h>
#include <cbm.h>
#include <errno.h>

#include "petsciipaint.h"

unsigned int vic = 0xD000; /* VIC-II control registers start address */
unsigned int bank = 0x8000; /* video bank start address */

unsigned int screens = 0xA000; /* first address seen by the VIC-II bitmap mode when bank 2 is selected and the bitmap is offset by 0x2000 bytes from bank start (0x8000) */
unsigned int cols = 0x8400; /* first address of the color information bytes when bank 2 is selected for the VIC-II chip */
unsigned int canvcols = 0x8428; /* first color info byte on canvas area */
unsigned int canvbits = 0xA140; /* first bit pattern byte on canvas area */

unsigned int spritepointers = 0x87F8; /* first sprite pointer in bank 2 */
unsigned int sp0 = 0x8000; /* sprite data start if sprite pointer is set to 0 */
unsigned int sp1 = 0x8040; /* sprite data start if sprite pointer is set to 1 */
unsigned int sp2 = 0x8080; /* sprite data start if sprite pointer is set to 2 */
unsigned int sp3 = 0x80C0; /* sprite data start if sprite pointer is set to 3 */
/* sprite info bytes start at: bank number * 16384 + spritepointer value * 64 */

/* sprite data for cursor (four frames of walking ants for an 8x8 size sprite) */
/* these values only describe the first eight rows of the first of three columns
   in each spritedata (everything else will be set to 0) */
unsigned char spdata0[] = { 204, 1, 1, 128, 128, 1, 1, 204 };                   
unsigned char spdata1[] = { 102, 0, 129, 129, 0, 0, 129, 153 };                     
unsigned char spdata2[] = { 51, 128, 128, 1, 1, 128, 128, 51 };
unsigned char spdata3[] = { 153, 129, 0, 0, 129, 129, 0, 102 };

unsigned char* charset; /* will get a copy of ROM chars from both C64 chargen charsets */
unsigned int charCount = 158; /* total number of selectable characters */

unsigned char* screenchars; /* for storing character codes for the 40*24 paint area */
unsigned char* screencolors; /* for storing a copy of the color information in the paint area */
unsigned char* brushes; /* will hold 8 brushes, 2 bytes each (one for char, one for colors) */

unsigned char mode=0; /* 0 edit-canvas mode, 1 brush-select mode, 2 enter-file-name mode */
signed int pcx = 0; /* cursor x pos before mode change */
signed int pcy = 8; /* cursor y pos before mode change */

unsigned char sb=0; /* selected brush */
unsigned char brushCount = 8;
unsigned char penstate=0; /* 0 pen is up, 1 pen is down */

unsigned char cf = 0; /* cursor frame */
unsigned char cc = 0; /* cursor color */
unsigned char cv = 1; /* cursor visibility */
signed int scrcx = 24; /* actual on screen cursor x pos */
signed int scrcy = 50; /* actual on screen cursor y pos */

signed int cx = 0; /* cursor x pos relative to first visible pixel column on the left*/
signed int cy = 8; /* cursor y pos relative to first visible pixel row on the top*/

unsigned char* filename = "s:unnamed  .petpnt,s";
/*                         0123456789112345678  */
/*                    SCRATCH        0      SEQ */
/* user entry routine may change chars from index 2 to 10 */
/* for load/save string will be used from index 2 to end */
/* for delete string will be used from index 0 to end */

void initbmpmode(void)
{
  /* change video bank to bank number 2
     (this will put the first address seen by the VIC-II to 0x8000)
     (the screen seen by the VIC-II can be set to start right there or at an offset of 0x2000 from there
     (I needed to change the bank and offset, so that my mallocs don't fall into the
      pretty unflexible screen byte ranges which would cause values in there to be shown
      as seemingly random garbage on the screen)
  */
  POKE(0xDD02, PEEK(0xDD02)|3); /* set bits 0 and 1 of port A of CIA#2 to 1(=output) */
  POKE(0xDD00, (PEEK(0xD00)&252)|1); /* select bank 2 (they are numbered 3,2,1,0 and selected by index in that list starting from 0 (e.g. bank 0 would be selected by putting a three into the control bits 3) */

  POKE(0xD018,PEEK(0xD018)|8); /* offset bitmap from bank-start (0x8000 for bank 2) by 0x2000 bytes (otherwise bitmap would collide with color info bytes) */
  POKE(0xD011,PEEK(0xD011)|32); /* activate bitmap mode */
}

void clearscreen(void)
{ 
  memset((void*)screens, 0, 8000);
  memset((void*)cols, 16, 1000); /* set default color to black and white for each 8x8 block */
  /* 0001 0000 bin is 16 dec (the first four bits set foreground color and the last four bits set background color) */
}

/* for brush-select mode */
void startbrushselectmode(void)
{
  unsigned char cBrush;
  unsigned int x,y;

  /* store last canvas mode cursor position */
  hidecursor();
  pcx = cx;
  pcy = cy;

  /* draw the charset to display selectable brushes 
     and change the color info to match colors for current brush */
  for(x=0; x<charCount; x++)
  {
    /* set screen color info byte to brush color byte */
    POKE(canvcols+x, PEEK(brushes+sb*2+1));

    /* draw the character bit pattern */
    for(y=0; y<8; y++)
      POKE(canvbits+x*8+y, PEEK(charset+x*8+y));   
  }

  /* empty the two empty spaces after the charset */
  blitchar(' ', 38, 4); blitchar(' ', 39, 4);
  colorchar(16, 38, 4); colorchar(16, 39, 4);

  /* draw a vertical line under the charset */
  for(x=0; x<40; x++)
  {
    blitchar( 68, x, 5);
    colorchar( 16, x, 5);
  }

  /* place cursor over currently selected brush */
  cBrush = PEEK(brushes+sb*2);
  cx = (cBrush % 40) * 8;
  cy = 8 + 8 * (cBrush / 40);

  updatecursor();

  mode = 1;

  showcursor();
}

void changeselectedbrush(unsigned char newchar)
{
  POKE(brushes+sb*2, newchar); /* store in brush info */
  blitchar(PEEK(brushes+sb*2), sb*4+2, 0); /* update status bar */
}

void endbrushselectmode(void)
{
  unsigned int i;
  unsigned int j;

  hidecursor();

  /* restore the canvas under the previously selectable brush area */
  for(j=1; j<6; j++) /* atm 4 rows of characters plus a row with a vertical line starting with the second row */
    for(i=0; i<40; i++)
    {
      blitchar(PEEK(screenchars+(j-1)*40+i), i, j);
      colorchar(PEEK(screencolors+(j-1)*40+i), i, j);
    }

  mode = 0;

  /* restore canvas mode cursor position */
  cx = pcx;
  cy = pcy;
  updatecursor();
  showcursor();
}

void blitchar(unsigned char c, unsigned char column, unsigned char row)
{
  unsigned int targetByte = screens+row*320+column*8;
  unsigned int sourceByte = (unsigned int)charset+c*8;
  unsigned char b;

  if (column >=40 || row >= 25 || c >= charCount) /* can skip < 0 check, because they are unsigned */
    return;

  for(b=0; b<8; b++)
    POKE(targetByte+b, PEEK(sourceByte+b));  
}

void colorchar(unsigned char colors, unsigned char column, unsigned char row)
{
  unsigned int targetByte = row*40+column;

  POKE(cols+targetByte, colors);
}

/* cmode 0 recolor character under cursor to brush color
   cmode 1 change character under cursor to brush character
   cmode 2 change color and character under cursor to brush character
*/
void drawatcursor(unsigned char cmode)
{
  unsigned char column = cx / 8;
  unsigned char row = cy / 8;

  /* update on screen information */
  if(cmode > 0)
    blitchar(PEEK(brushes+sb*2), column, row);

  if(cmode != 1)
    colorchar(PEEK(brushes+sb*2+1), column, row);

  /* update internal buffers */
  if(cmode > 0)
    POKE(screenchars+(row-1)*40+column, PEEK(brushes+sb*2));

  if(cmode != 1)
    POKE(screencolors+(row-1)*40+column, PEEK(brushes+sb*2+1)); 
}

/* cmode 0 copy color under cursor into current brush
   cmode 1 copy character under cursor into current brush
   cmode 2 copy color and character under cursor into current brush 
*/
void copyfromcursortobrush(unsigned char cmode)
{
  unsigned char column = cx / 8;
  unsigned char row = cy / 8;

  if(cmode > 0)
  {
    POKE(brushes+sb*2, PEEK(screenchars+(row-1)*40+column)); /* grab character from character buffer into brush */
    blitchar( PEEK(brushes+sb*2), sb*4+2, 0 ); /* update status bar brush character */
  }

  if(cmode != 1)
  {
    unsigned char col = PEEK(screencolors+(row-1)*40+column);
    POKE(brushes+sb*2+1, col); /* grab color from color buffer into brush info */
    POKE(cols+sb*4+1, col); /* update status bar brush color wedge */
    POKE(cols+sb*4+2, col); /* update status bar brush character color */
  }
}

void clearstatusbar(void)
{
  memset((void*)screens, 0, 320); /* clear bitpatterns in first 40x8 bytes (row 0) */
  memset((void*)cols, 16, 40); /* set color information for first 40 8x8 blocks to black and white */
}

void drawpenstatus(void)
{
  if (penstate==0)
    blitchar(45, 39, 0); /* - in upper right corner */
  else
    blitchar(43, 39, 0); /* + in upper right corner */
}

void drawstatusbar(void)
{
  unsigned char b;
  unsigned int boff;
  unsigned char b4;

  clearstatusbar();

  /* draw brush information */
  for(b=0; b<brushCount; b++)
  {
    boff = (unsigned int)brushes+b*2;
    b4 = b*4;

    blitchar(49+b, b4, 0); /* brush number */
    blitchar(PEEK(boff), b4+2, 0); /* the actual brush character */
    blitchar(105, b4+1, 0); /* upper left wedge for color information */
    colorchar(PEEK(boff+1), b4+1, 0); /* color the color information wedge */
    colorchar(PEEK(boff+1), b4+2, 0); /* color the brush character itself as it will be placed on the canvas */

    if(b!=sb) /* color the brush number if the brush is not selected */
      colorchar( 176, b*4, 0); /* 11 for foreground (gray1), 0 for background (black) */
  }

  drawpenstatus();
}

void selectbrush(unsigned char b)
{
  if (b>=8)
    return;

  colorchar( 176, sb*4, 0); /* gray out currently selected brush number */
  sb = b;
  colorchar(  16, sb*4, 0); /* highlight new selected brush number */

  if (mode == 1) /* brush-select mode */
  {
    unsigned int boff = (unsigned int)brushes + sb*2;

    /* place cursor over new selected brush */
    cx = (PEEK(boff) % 40) * 8;
    cy = 8 + 8 * (PEEK(boff) / 40);

    updatecursor();
    
    /* recolor displayed charset to new selected brush colors */
    memset((void*)canvcols, PEEK(boff+1), charCount);
  }
}

/* direction 0 selects previous color, 1 selects next color, b is the brush number 0to7 
   fgorbg 0 change foreground, 1 change background
*/
void updatebrushcolor(unsigned char b, unsigned char direction, unsigned char fgorbg)
{
  unsigned char bg;
  unsigned char fg;
  unsigned int boff = (unsigned int)brushes+b*2+1;

  if (b>=8)
    return;
  
  bg = PEEK(boff)&15; /* store background nybble */
  fg = (PEEK(boff)&240)>>4; /* store foreground nybble */

  /* change color */
  if(direction == 0)
  {
    if (fgorbg == 0)
    {
      if(fg < 1)
        fg = 15;
      else
        fg = fg-1;
    }
    else
    {
      if(bg < 1)
        bg = 15;
      else
        bg = bg-1;
    }
  }
  else
  {
    if (fgorbg == 0)
    {
      if(fg > 14)
       fg = 0;
      else
       fg = fg+1;
    }
    else
    {
      if(bg > 14)
       bg = 0;
      else
       bg = bg+1;
    }
  }  

  fg = fg<<4;

  /* store new color information for brush */
  POKE(boff, fg|bg);

  /* update on screen color information character (next to the brush number) for brush */
  colorchar( fg|bg, b*4+1, 0);
  
  /* also color the brush character itself so it displays exactly what will be placed on the canvas */
  colorchar( fg|bg, b*4+2, 0);
 
  if(mode == 1) /* brush-select mode */
    memset((void*)canvcols, PEEK(brushes+sb*2+1), charCount); /* set screen color info bytes to brush color byte */
}

void enablecursor(void)
{
  /* initial cursor position relative to upper left corner of visible screen area */
  cx = 0;
  cy = 8;

  POKE(vic+39, 1); /* set sprite 0 color to 1 (white) */

  POKE(vic, cx); /* set x position to 8 for sprite 0 */
  POKE(vic+16, 0); /* clear most significant bit for all sprites */
  POKE(vic+1, cy); /* set y position to 8 for sprite 0 */

  POKE(vic+21, 1); /* activate sprite 0, deactivate all others */
}

void hidecursor(void)
{
  POKE(vic+21, 0);
}

void showcursor(void)
{
  POKE(vic+21, 1);
}

void togglecursor(void)
{
  cv ^= 1;

  POKE(vic+21,cv);
} 

void togglepenstate(void)
{
  penstate ^= 1;

  drawpenstatus();
}

void updatecursor(void)
{
  unsigned int tb; /* theoretically selected brush character in brush-select mode */

  /* update cursor color and frame */
  cc+=1;
  if (cc > 15)
    cc = 0;
  cf+=1;
  if(cf >3)
    cf=0;

  /* set cursor color (sprite 0 color) */
  POKE(vic+39, cc);

  /* set cursor frame (sprite 0 pointer) */
  POKE(spritepointers, cf);

  if(mode == 0) /* sanitize and update cursor position in edit-canvas mode */
  {
    if (cx >= 320)
      cx = cx-320;
    if (cx <0)
      cx = 320+cx;
    if (cy >= 200)
      cy = cy-200+8; /* cursor is not allowed on the first line, hence the 8 */
    if (cy <8)
      cy = 200-8+cy;
  }

  if(mode == 1) /* sanitize and update cursor position in brush-select mode */
  {
    if (cx >= 320)
      cx = cx-320;
    if (cx <0)
      cx = 320+cx;
    if (cy >= 40) /* in this mode the cursor can't go beyond line 4 */
      cy = cy-40+8; /* cursor is not allowed on the first line, hence the 8 */
    if (cy <8)
      cy = 40-8+cy;

    tb = ((cy-8)/8)*40+(cx/8); /* determine theoretically selected char */

    if(tb >= charCount) /* selection invalid at that position? need to adjust cursor */
    {
      tb = tb-charCount; /* wrap around characters */
      cx = (tb % 40) * 8;
      cy = 8 + (tb / 40) * 8;
    }

    changeselectedbrush(tb);
  }

  if(mode == 2) /* sanitize cursor position in enter-file-name mode */
  {
    cy = 0; /* cursor is always in first row in this mode */
    if (cx >= 320)
      cx = 0;
    if (cx < 0)
      cx = 0;
  }

  scrcx = cx + 24;
  scrcy = cy + 50;

  if(scrcx < 256)
  {
    POKE(vic, scrcx);
    POKE(vic+16,0);
  }
  else
  {
    POKE(vic, scrcx-256);
    POKE(vic+16, 1);
  }
  
  POKE(vic+1, scrcy);
}

/* cmode 0 init canvas memory to defaults, 1 like 0 and update screen accordingly */
/* cmode 2 init canvas memory to current brush colors and update screen */
void clearcanvas(unsigned char cmode)
{
  unsigned char col = cmode == 0 ? 16 : PEEK(brushes+sb*2+1);

  /* update internal buffers */
  memset(screenchars, 32, 40*24); /* set everything to "space" char */
  memset(screencolors, col, 40*24); /* set color buffer */  

  if(cmode > 0)
  {
    memset((void*)canvbits, 0, 7680); /* clear the canvas area of the screen */
    memset((void*)canvcols, col, 960); /* set color info in canvas area of the screen */
  }
}

void delayNOP(unsigned int count)
{
  unsigned int d;
  for (d=0; d<count; d++)
  {
    __asm__("nop");
  }
}

void allocateram(void)
{
  charset = malloc(charCount*8);
  screenchars = malloc(40*24);
  screencolors = malloc(40*24);
  brushes = malloc(brushCount*2);
  clearcanvas(0);
}

void freeram(void)
{
  free(brushes);
  free(screencolors);
  free(screenchars);
  free(charset); 
}

/*
void displaypointers(void)
{
  printf("charset: %x\n", charset);
  printf("screenchars: %x\n", screenchars);
  printf("screencolors: %x\n", screencolors);
}
*/

void initbrushes(void)
{
  unsigned char b;
  
  sb = 0;

  for(b=0; b<brushCount; b++)
  {
    POKE(brushes+b*2+1, 16); /* set black and white as default colors */
  }
  
  /* set default characters for brushes */
  POKE(brushes   ,  81); /* round ball */
  POKE(brushes+2 ,  87); /* hollow ball */
  POKE(brushes+4 , 105); /* upper left wedge */
  POKE(brushes+6 ,  95); /* upper right wedge */
  POKE(brushes+8 ,  65); /* spades */
  POKE(brushes+10,  83); /* hearts */
  POKE(brushes+12,  88); /* clubs */
  POKE(brushes+14,  90); /* diamonds */
}

void initcharset(void)
{
  unsigned int x;
  unsigned int s;

  /* turn off interrupt */
  POKE(0xDC0E, PEEK(0xDC0E)&254);

  /* disable IO and enable ROM-chargen
     which will mirror the character bit patterns at a fixed RAM address (0xD000) */
  POKE(0x0001, PEEK(0x0001)&251);

  /* copy character bitpatterns from chargen characters 
     temporarily mirrored in RAM by ROM-chargen
     to own allocated charset RAM */
  for(x=0; x<128*8; x++) /* copy the first 128 characters from charset 1 */
  {
    POKE(charset+x, PEEK(0xD000+x));
  }

  /* poke in mr. smiley face to character number 128 */
  POKE(charset+1024, 124); /* 01111100 */
  POKE(charset+1025, 130); /* 10000010 */
  POKE(charset+1026, 170); /* 10101010 */
  POKE(charset+1027, 130); /* 10000010 */
  POKE(charset+1028, 186); /* 10111010 */
  POKE(charset+1029, 146); /* 10010010 */
  POKE(charset+1030, 130); /* 10000010 */
  POKE(charset+1031, 124); /* 01111100 */

  for(x=129*8; x<155*8; x++) /* copy characters 1 to 26 (lowercase letters) from charset 2 */
  {
    POKE(charset+x, PEEK(0xD400+x));
  }
  /* copy the remaining unique characters from charset 2 which are not in charset 1 already */
  s = 155*8;
  for(x=0; x<8; x++) /* copy character 95 from charset 2 */
    POKE(charset+s+x, PEEK(0xD800+95*8+x));
  s = 156*8;
  for(x=0; x<8; x++) /* copy character 105 from charset 2 */
    POKE(charset+s+x, PEEK(0xD800+105*8+x));
  s = 157*8;
  for(x=0; x<8; x++) /* copy character 122 from charset 2 */
    POKE(charset+s+x, PEEK(0xD800+122*8+x));

  /* enable IO and disable ROM-chargen
     which will stop mirroring the character bit patters at the fixed RAM address */
  POKE(0x0001, PEEK(0x0001)|4);

  /* turn on interrupt */
  POKE(0xDC0E, PEEK(0xDC0E)|1);
}

unsigned char stringconsttomychar(unsigned char scchar)
{
  if(scchar >= ' ' && scchar <= '?')
    return scchar;

  if(scchar >= 'a' && scchar <= 'z')
    return 154 + scchar - 'z';

  if(scchar >= 'A' && scchar <= 'Z')
    return 26 + scchar - 'Z';

  if(scchar == '[')
    return 27;
  if(scchar == ']')
    return 29;

  return 63; /* return ? for every unknown/unsupported char */
}

void printstring(unsigned char* s, unsigned char column, unsigned char row, unsigned char maxCharsToPrint)
{
  unsigned char index=0;
  unsigned char curcol = column;
  unsigned char currow = row;

  /* print only as long as next char exists and is within screen boundaries */  
  while(s[index]!=0 && index < maxCharsToPrint && curcol < 40 && currow < 25)
  {
    blitchar(stringconsttomychar(s[index]), curcol, currow);
   
    curcol += 1;
    if(curcol > 39)
    {
      curcol = 0;
      currow += 1;
    }
    index++;
  }
}

void initspritedata(void)
{
  unsigned int i;

  /* init sprite pointer to point to sp0 for sprite 0 */
  POKE(spritepointers, 0);

  /* clear sprite data */
  memset((void*)sp0, 0, 64);
  memset((void*)sp1, 0, 64);
  memset((void*)sp2, 0, 64);
  memset((void*)sp3, 0, 64);

  /* set sprite data, skipping the second and third byte for each row */
  for(i=0; i<8; i++)
  {
    POKE(sp0+i*3, PEEK(spdata0+i));
    POKE(sp1+i*3, PEEK(spdata1+i));
    POKE(sp2+i*3, PEEK(spdata2+i));
    POKE(sp3+i*3, PEEK(spdata3+i));
  }
}

void shutbmpmode(void)
{
  unsigned int c;

  POKE(0xD011,PEEK(0xD011)&223); /* de-activate bitmap mode */

  /* select default video bank */
  POKE(0xDD02, PEEK(0xDD02)|3); /* set bits 0 and 1 of port A of CIA#2 to 1(=output) */
  POKE(0xDD00, (PEEK(0xD00)&252)|3); /* select bank 0 (they are numbered 3,2,1,0 and selected by index in that list starting from 0 (e.g. bank 0 would be selected by putting a three into the control bits 3) */

  /* set default basic colors */
  POKE(646, 14);
  POKE(0xD020,14);
  POKE(0xD021,6);

  /* call KERNAL routines RESTOR and CINT */
  __asm__("jsr $FF8A"); /* restores input/output vector to default */
  __asm__("jsr $FF81"); /* initializes the basic screen editor */

  /* reset the color information RAM */
  for(c=0; c<1000; c++)
    POKE(0xD800+c, 6);

  /* clear the screen */
  printf("%cthank you for using petscii paint\nhttp://www.dennisbusch.de\n", 147);
}

void init(void)
{
  allocateram();

  initbmpmode();

  initcharset();

  initbrushes();

  initspritedata();

  clearscreen();

  enablecursor();
}

void shut(void)
{
  hidecursor();

  shutbmpmode();

  freeram();
}

/* fmode 0 for save, 1 for load */
/* returns 0 on ENTER, 1 on ABORT */
unsigned char enterfilename(unsigned char fmode)
{
  unsigned char fnI; /* current filename character index */
  unsigned char i;
  unsigned char key = 0;
  unsigned char mincx = 17*8;

  if(fmode == 0)
    printstring("SAVE, FILENAME: [         ] (* CANCELS)", 0, 0, 40);
//               01234567890123456789012345678901
//                      min cx=17*8       max cx=25*8
  if(fmode == 1)
    printstring("LOAD, FILENAME: [         ] (* CANCELS)", 0, 0, 40);

  /* find current index (last space char in user settable filename part) */
  fnI = 10;
  for(i=10; i>1; i--)
  {
    if(filename[i] == ' ')
      fnI = i;
    else
      break;
  }

  cx = mincx + (fnI-2)*8;
  updatecursor();
  showcursor();

  while(key != 13 && key != '*') /* 13 is ENTER, * is for abort */
  {
    if(kbhit())
      key = cgetc();
 
    if(key == 20 && fnI > 2) /* 20 is backspace */
    {
      if (fnI == 10 && filename[fnI] != ' ')
      {
        filename[fnI] = ' ';
      }
      else
      {
        filename[fnI] = ' ';
        fnI--;
        filename[fnI] = ' ';
      }

      key = 0;
    }

    /* allow only valid filename characters */
    if(key != 0)
    {
      unsigned char validkey = 0;

      if(key >= 'a' && key <= 'z')
      {
        filename[fnI] = key;
        validkey = 1;
      }

      if(key >= 'A' && key <= 'Z')
      {
        filename[fnI] = 'z' + key - 'Z';
        validkey = 1;
      }

      if((key >= '0' && key <= '9') || key == ' ')
      {
        filename[fnI] = key;
        validkey = 1;
      }
    
      if(fnI < 10 && validkey == 1)
        fnI++;
    }

    if(key != 13 && key != '*')
      key = 0;

    cx = mincx + (fnI-2)*8;
    updatecursor();
    delayNOP(800);

    printstring(&filename[2], 17, 0, 9);
  }

  if(key == 13)
    return 0;
  else
    return 1;
}

void startloadsave(void)
{
  hidecursor();
  pcx = cx;
  pcy = cy;
  clearstatusbar();
  mode = 2;
}

void endloadsave(void)
{
  hidecursor();
  drawstatusbar();
  cx = pcx;
  cy = pcy;
  mode = 0;
  updatecursor();
  showcursor();
}

/* returns 1 if the currently set file in filename (starting at index 2) exists */
unsigned char fileexists(void)
{
  unsigned char* testread = "00";
  unsigned char readbytes = 0;
  unsigned char opencode = 0;
  
  opencode = cbm_open(1,8,0, &filename[2]);
  if(opencode == 0)
  {     
    readbytes = cbm_read(1, (void*)&testread, 1); 
  }
  cbm_close(1);
  
  if(readbytes > 0)
    return 1;
  else
    return 0;
}

void deletecurrentfile(void)
{
  clearstatusbar();
  printstring("DELETING FILE: ", 0, 0, 20);
  printstring(&filename[2], 15, 0, 9);

  cbm_open(1, 8, 15, filename); /* the S: for scratching is always at the start of that string */
  cbm_close(1);
}

/* returns 1 if saving was successfull */
unsigned char savecanvastofile(void)
{
  unsigned char opencode = 0;
  signed int byteswritten = 0;

  hidecursor();

  opencode = cbm_open(1,8,1,&filename[2]);
  if(opencode == 0)
  {
    clearstatusbar();
    printstring("saving brushes...", 0, 0, 20);
    byteswritten = cbm_write(1, (void*)brushes, brushCount*2);
    if(byteswritten != brushCount*2)
    {
      cbm_close(1);
      return 0;
    }

    clearstatusbar();
    printstring("saving chars...", 0, 0, 20);
    byteswritten = cbm_write(1, (void*)screenchars, 40*24);
    if(byteswritten != 40*24)
    {
      cbm_close(1);
      return 0;
    }
    
    clearstatusbar();
    printstring("saving colors...", 0, 0, 20);
    byteswritten = cbm_write(1, (void*)screencolors, 40*24);
    if(byteswritten != 40*24)
    {
      cbm_close(1);
      return 0;
    }

    cbm_close(1);
    return 1;
  }
  else
  {
    cbm_close(1);
    return 0;
  }
}

/* returns 1 if loading was successfull */
unsigned char loadcanvasfromfile(void)
{
  unsigned char opencode = 0;
  signed int bytesread = 0;
 
  hidecursor();

  opencode = cbm_open(1, 8, 0, &filename[2]);
  if(opencode == 0)
  {
    clearstatusbar();
    printstring("loading brushes...", 0, 0, 20);
    bytesread = cbm_read(1, (void*)brushes, brushCount*2);
    if(bytesread != brushCount*2)
    {
      cbm_close(1);
      return 0;
    }

    clearstatusbar();
    printstring("loading chars...", 0, 0, 20);
    bytesread = cbm_read(1, (void*)screenchars, 40*24);
    if(bytesread != 40*24)
    {
      cbm_close(1);
      return 0;
    }
    
    clearstatusbar();
    printstring("loading colors...", 0, 0, 20);
    bytesread = cbm_read(1, (void*)screencolors, 40*24);
    if(bytesread != 40*24)
    {
      cbm_close(1);
      return 0;
    }

    cbm_close(1);
    return 1;
  }
  else
  {
    cbm_close(1);
    return 0;
  }
}

void savefile(void)
{
  unsigned char failed = 0;
  unsigned char key = 0;
  unsigned char proceed = 1;
  startloadsave();

  if( enterfilename(0) == 0) /* ENTER */
  {
    if (fileexists() == 1)
    {
      clearstatusbar();
      printstring("overwrite existing file? (y/n)", 0, 0, 40);
      
      while(key != 'y' && key != 'n')
      {
        key = cgetc();
      }

      if (key == 'n')
        proceed = 0;
      else
        deletecurrentfile();
    }

    if(proceed == 1)
    {
      if( savecanvastofile() == 1 )
      {
        clearstatusbar();
        printstring("saved successfully. (press spacebar)", 0, 0, 40);
      }
      else
      {
        clearstatusbar();
        printstring("saving failed! (press spacebar)", 0, 0, 40);
      }
      
      while(cgetc() != 32);
    }
  }

  endloadsave();
}

void loadfile(void)
{
  unsigned char failed = 0;
  startloadsave();

  if( enterfilename(1) == 0) /* ENTER */
  {
    if(fileexists() == 1)
    {
      if(loadcanvasfromfile() == 1)
      {
        drawcanvas();
      }
      else
        failed = 1;
    }
    else
      failed = 1;

    if(failed)
    {
      clearstatusbar();
      printstring("loading failed! (press spacebar)", 0, 0, 40);
      while(cgetc() != 32);
    }
  }

  endloadsave();
}

void showhelpscreen(void)
{
  hidecursor();
  clearscreen();

  printstring("PETSCII PAINT Help Screen", 0, 0, 40);
  printstring("http://www.dennisbusch.de", 0, 1, 40);

  printstring("F1 to open/close this help screen", 0, 3, 40);
  printstring("1 to 8 to select brush", 0, 4,40);
  printstring("b to open/close character select mode", 0, 5,40);
  printstring("i,j,k,l to move cursor around", 0, 7,40);
  printstring("(hold SHIFT to move faster)", 0, 8,40);
  printstring("a,s to cycle foreground color for brush", 0, 9,40);
  printstring("n,m to cycle background color for brush", 0, 10,40);
  printstring("q,w,e grab color,char,both from canvas", 0, 11,40);
  printstring("d,f,r place color,char,both on canvas", 0, 12,40);
  printstring("p toggles pen status", 0, 13,40);
  printstring("h toggles cursor visibility", 0, 14,40);
  printstring("SHIFT+C clear canvas to brush colors", 0, 15, 40);
  
  printstring("F3 to save to disk in device 8", 0, 17,40);
  printstring("F5 to load from disk in device 8", 0, 18,40);
  printstring("SHIFT+X to end PETSCII PAINT", 0, 19,40);
  
  while(cgetc() != 133); /* wait for F1 */

  drawcanvas();
  drawstatusbar();
  showcursor();
}

void drawcanvas(void)
{
  unsigned char col;
  unsigned char row;

  clearstatusbar();
  printstring("restoring canvas...", 0, 0, 40);

  for(row=1; row<25; row++)
    for(col=0; col<40; col++)
    {
      blitchar(PEEK(screenchars+(row-1)*40+col), col, row);
      colorchar(PEEK(screencolors+(row-1)*40+col), col, row);
    }
}

/*
void inputtest(void)
{
  unsigned char c = 0;
  while(c != 3) 
  {
   c = cgetc();
   printf("%i\n", c);
  }
}
*/

int main(void) 
{ 
  unsigned char exit = 0;
  unsigned char fastmovedir = 0; /* 0 no fast move, 1 right, 2 left , 3 down, 4 up */
  unsigned char x = 0;

  init();
  drawstatusbar();

  while(exit == 0)
  {
    delayNOP( (mode==0) ? 800 : 700); /* delay is adjusted to heavier CPU load in brush select mode */

    if(kbhit())
    {
      x = cgetc();
      // while(kbhit());
    }
    
    /* cursor movement */  
    if(x == 'l')
      cx += 8;
    if(x == 'j')
      cx -= 8;
    if(x == 'k')
      cy += 8;
    if(x == 'i')
      cy -= 8;

    updatecursor(); /* sanitize cursor position and update its' display sprite */

    /* color cycling for selected brush */
      if (x == 'a')
        updatebrushcolor(sb, 0, 0);
      if (x == 's')
        updatebrushcolor(sb, 1, 0);
      if (x == 'n')
        updatebrushcolor(sb, 0, 1);
      if (x == 'm')
        updatebrushcolor(sb, 1, 1);

    /* brush selection */
      if (x > 48 && x < 57) /* number characters 1 to 8 */
        selectbrush(7-(56-x));

    if(mode == 0) /* edit canvas mode exclusive keys */
    {
      /* help screen */
      if(x == 133) /* F1 */
        showhelpscreen();

      /* saving and loading of canvas */
      if(x == 134) /* F3 */
        savefile();
      if(x == 135) /* F5 */
        loadfile();

      /* cursor visibility */
      if(x == 'h')
        togglecursor();

      /* penstate */
      if(x == 'p')
        togglepenstate();

      /* fast cursor movement */
      fastmovedir = 0;
      if(x == 'L')
        fastmovedir = 1;
      if(x == 'J')
        fastmovedir = 2;
      if(x == 'K')
        fastmovedir = 3;
      if(x == 'I')
        fastmovedir = 4;

      /* drawing and brush operations */
      if (x == 'd')
        drawatcursor(0);
      if (x == 'f')
        drawatcursor(1);
      if (x == 'r')
        drawatcursor(2);

      if (x == 'q')
        copyfromcursortobrush(0);
      if (x == 'w')
        copyfromcursortobrush(1);
      if (x == 'e')
        copyfromcursortobrush(2);

      if (x == 'C')
        clearcanvas(2);

      if (penstate == 1)
        drawatcursor(2);

      /* fast move evaluation */
      if(fastmovedir > 0)
      {
        unsigned char i;
        for(i=0; i<5; i++)
        {
          if (fastmovedir==1)
            cx+=8;
          if (fastmovedir==2)
            cx-=8;
          if (fastmovedir==3)
            cy+=8;
          if (fastmovedir==4)
            cy-=8;

          updatecursor();

          if(penstate == 1)
            drawatcursor(2);  
        }
      }

      /* mode changes */
      if (x == 'b')
      {
        startbrushselectmode();
        x = 0; /* clear key so mode does not immediately exit again */
      }
    }

    if(mode == 1) /* brush select mode exclusive keys */
    {
      /* mode changes */
      if (x == 'b')
      {
        endbrushselectmode();
      }
    }

    if(x == 'X') /* shift+x to exit */
      exit = 1;

    x=0; /* clear last key */
  }
  shut();

  // debugging stuff
  //allocateram();
  //displaypointers();
  //freeram();
  //inputtest();

  return EXIT_SUCCESS; 
}

